from .base import ConfigStoreBase, ConfigurationNotFound


class RedisStore(ConfigStoreBase):
    __slots__ = ["parser", "redis"]

    def __init__(self, parser, arguments):
        """Initialize the Redis configuration store

        :param parser: Configuration file/data parser
        :type parser: minty.config.parser.Base
        :param redis: dictionary containing keys/values to configure a Redis
            client instance.
        :type arguments: dict
        """

        self.parser = parser
        self.redis = arguments

    def retrieve(self, name) -> str:
        """Retrieve configuration from Redis

        :param name: Name of the configuration to retrieve
        :param name: str
        :raises ConfigurationNotFound: if the named configuration can't be found
        :return: The configuration "file" contents
        :rtype: str
        """

        super().retrieve(name)

        with self.statsd.get_timer().time("retrieve_time"):
            config = self.redis.get("saas:instance:" + name)

        self.statsd.get_counter().increment("retrieve")

        if config is None:
            self.statsd.get_counter().increment("retrieve_miss")
            raise ConfigurationNotFound

        parsed = self.parser.parse(config.decode("utf-8"))

        return parsed["instance"]



from .base import ConfigParserBase
from apacheconfig import make_loader


class ApacheConfigParser(ConfigParserBase):
    __slots__ = ["content"]

    def parse(self, content: str) -> dict:
        """Return a dict with the parsed config.

        :return: Parsed configuration
        :rtype: dict
        """
        content = super().parse(content)

        if content == "":
            raise ValueError("Cannot parse empty configuration")

        # statsd configuration needs to be loaded before it can be used,
        # so we can't measure the loading easily
        with make_loader() as loader:
            config = loader.loads(content)

        if "statsd" in config:
            self.statsd.get_counter().increment("parse")

        return config



import json
from .base import ConfigParserBase


class JSONConfigParser(ConfigParserBase):
    __slots__ = ["content"]

    def parse(self, content: str) -> dict:
        """Parse JSON file and return dict.

        :param content: content from JSON file
        :type content: str
        :raises ValueError: if content is empty
        :return: parsed config dict
        :rtype: dict
        """
        content = super().parse(content)

        if content == "":
            raise ValueError("Cannot parse empty configuration")

        with self.statsd.get_timer().time("parse_time"):
            config = json.loads(content)
        self.statsd.get_counter().increment("parse")
        return config



from pydantic import BaseModel, Field, ValidationError
from pydantic.generics import GenericModel

__all__ = ["IntrospectableObject", "Field", "ValidationError", "GenericModel"]


class IntrospectableObject(BaseModel):
    """ Introspectable object based on pydantic """

    class Config:
        validate_assignment = True



import holidays
from datetime import date, datetime, timedelta
from dateutil.relativedelta import relativedelta
from dateutil.tz import gettz


def add_timedelta_to_date(
    interval_type: str, time: int, starting_date: date = None
) -> date:
    """Add timedelta to given date.

    :param interval_type: type of interval
    :type interval_type: str
    :param time: time to add to date
    :type time: int
    :param starting_date: date to add time to, defaults to None
    :type starting_date: date, optional
    :raises KeyError: when interval_type does not exist.
    :return: calculated end date
    :rtype: datetime.date
    """
    if starting_date is None:
        starting_date = datetime.now(tz=gettz("Europe/Amsterdam")).date()

    if interval_type == "days":
        end_date = starting_date + timedelta(days=time)
    elif interval_type == "business_days":
        nwd = next_business_day(starting_date)
        for _ in range(time):
            end_date = next(nwd)
    elif interval_type == "weeks":
        end_date = starting_date + timedelta(weeks=time)
    elif interval_type == "months":
        end_date = starting_date + relativedelta(months=time)
    elif interval_type == "years":
        end_date = starting_date + relativedelta(years=time)
    else:
        raise KeyError(f"Interval type: '{interval_type}' does not exist.")

    return end_date


def next_business_day(initial_date: date) -> date:
    """Calculate next business day after given date.

    :param initial_date: date to calculate from.
    :type initial_date: date
    """
    non_business_days_store = NonBusinessDays()
    date_to_check = initial_date + timedelta(days=1)
    while True:
        if date_to_check.isoweekday() not in (6, 7):
            non_business_days = non_business_days_store(date_to_check.year)
            if date_to_check not in non_business_days:
                yield date_to_check
        date_to_check = date_to_check + timedelta(days=1)


class NonBusinessDays:
    def __init__(self):
        self.store = {}

    def __call__(self, year: int):
        """Get non business days by given year.

        Checks if given year exists in `self.store` else given year will be
        initialized and stored in `self.store` for next retrieval.

        :param year: year
        :type year: int
        :return: non business days for given year
        :rtype: dict
        """
        try:
            non_business_days = self.store[year]
        except KeyError:
            self.store[year] = holidays.Netherlands(years=year)
            non_business_days = self.store[year]
        return non_business_days



from minty.cqrs import EventService
from unittest import mock
from uuid import uuid4


class InfraFactoryMock:
    """Mock for our InfraFactory"""

    def __init__(self, infra):
        self.infra = infra

    def get_infrastructure(self, context, infrastructure_name):
        return self.infra[infrastructure_name]


class MockRepositoryFactory:
    """Mock for our repository factory"""

    def __init__(self, infra_factory):
        self.infrastructure_factory = infra_factory
        self.repositories = {}

    def get_repository(self, name, context, event_service):
        mock_repo = self.repositories[name]
        return mock_repo


class TestBase:
    """Utilities for making tests on our CQRS infrastructure without
    database access
    """

    def load_query_instance(
        self, domain: object, repositoryname: str, inframocks: dict = {}
    ):
        """ Loads your object with a qry attribute pointing to the query
        instance

        Will set the self.qry property to the loaded Queries() object,
        containing mocks for:
        * Database
        * S3

        Args:
            domain: Domain object to use in this test,
            e.g.: zsnl_domains.case_management
            repositoryname: The repository to use, e.g.: "case"
            inframocks: A key-value pair containing mocks for every component
            (e.g { "database": MagicMock })

        """

        mock_session = mock.MagicMock()
        self.qry = self._get_mocked_domain_instance(
            session_mock=mock_session,
            domain=domain,
            repositoryname=repositoryname,
            instance_type="query",
            inframocks=inframocks,
        )
        self.session = mock_session

    def load_command_instance(
        self, domain: object, repositoryname: str, inframocks: dict = {}
    ):
        """ Loads your object with a cmd attribute pointing to the command
        instance

        Will set the self.cmd property to the loaded Queries() object,
        containing mocks for:
        * Database
        * S3

        Args:
            domain: Domain object to use in this test,
                    e.g.: zsnl_domains.case_management
            repositoryname: The repository to use, e.g.: "case"
            inframocks: A key-value pair containing mocks for every component
                        (e.g { "database": MagicMock })

        """

        mock_session = mock.MagicMock()
        self.cmd = self._get_mocked_domain_instance(
            session_mock=mock_session,
            domain=domain,
            repositoryname=repositoryname,
            instance_type="command",
            inframocks=inframocks,
        )
        self.session = mock_session

    def assert_has_event_name(self, event_name: str, description: str = None):
        list_of_events = [
            ev
            for ev in self.cmd.event_service.event_list
            if ev.event_name == event_name
        ]

        if description:
            assert len(list_of_events) > 0, description
        else:
            assert len(list_of_events) > 0

        return True

    def get_last_sql_query(self):
        (args, kwargs) = self.session.execute.call_args
        query = args[0].compile()

        return query

    def _get_mocked_domain_instance(
        self,
        session_mock,
        domain,
        repositoryname,
        instance_type,
        inframocks={},
        user_uuid=None,
    ):
        """Generates a mock for a command or query instance"""

        # Some defaults
        mocked_infra = {
            "database": session_mock,
            "s3": mock.MagicMock(),
            **inframocks,
        }

        repository = domain.REQUIRED_REPOSITORIES[repositoryname]

        if not user_uuid:
            user_uuid = str(uuid4())

        event_service = EventService(
            correlation_id=str(uuid4()),
            domain="domain",
            context="context",
            user_uuid=user_uuid,
        )

        infra = InfraFactoryMock(infra=mocked_infra)
        repo_factory = MockRepositoryFactory(infra_factory=infra)
        repo_factory.repositories[repositoryname] = repository(
            infrastructure_factory=infra,
            context=None,
            event_service=event_service,
        )

        if instance_type == "query":
            return domain.get_query_instance(
                repository_factory=repo_factory,
                context=None,
                user_uuid=user_uuid,
            )
        elif instance_type == "command":
            return domain.get_command_instance(
                repository_factory=repo_factory,
                event_service=event_service,
                context=None,
                user_uuid=user_uuid,
            )



import wrapt
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from minty.object import Field, IntrospectableObject
from typing import List
from uuid import UUID


class EntityBase(ABC):
    @property
    @abstractmethod
    def entity_id(self):
        raise NotImplementedError

    @property
    def event_service(self):
        return self._event_service

    @event_service.setter
    def event_service(self, event_service):
        super().__setattr__("_event_service", event_service)

    @property
    def change_log(self):
        try:
            self._changes
        except AttributeError:
            self.clear_change_log()
        return self._changes

    def clear_change_log(self):
        super().__setattr__("_changes", [])

    @property
    def entity_data(self):
        try:
            self._entity_data
        except AttributeError:
            self.clear_entity_data()
        return self._entity_data

    def clear_entity_data(self):
        super().__setattr__("_entity_data", {})

    def __setattr__(self, attr, value):
        try:
            old_value = self.__getattribute__(attr)
            change = {
                "key": attr,
                "old_value": _reflect(old_value),
                "new_value": _reflect(value),
            }

            self.change_log.append(change)
        except AttributeError:
            pass
        super().__setattr__(attr, value)

    def capture_field_values(self, fields: List):
        for field in fields:
            value = self.__getattribute__(field)
            self.entity_data[field] = _reflect(value)


def _reflect(value):
    """Reflect on attribute type and return JSON parse-able type.

    :param attr: attribute
    :return: converted object to correct type
    :rtype: str, int or None
    """
    if (value is None) or isinstance(value, bool) or isinstance(value, int):
        return value

    if isinstance(value, EntityBase):
        return {
            "type": value.__class__.__name__,
            "entity_id": str(value.entity_id),
        }

    if isinstance(value, Entity):
        return {k: _reflect(v) for k, v in value.entity_dict().items()}
    elif isinstance(value, IntrospectableObject):
        return {k: _reflect(v) for k, v in value.dict().items()}

    if isinstance(value, List):
        return [_reflect(i) for i in value]

    if isinstance(value, Enum):
        return value.value

    if isinstance(value, dict):
        return {k: _reflect(v) for k, v in value.items()}

    if isinstance(value, datetime):
        return value.isoformat()

    return str(value)


class ValueObject(IntrospectableObject):
    pass


class Entity(IntrospectableObject):
    """ Pydantic based Entity object

    Entity object based on pydantic and the "new way" of creating entities in
    our minty platform. It has the same functionality as the EntityBase object,
    but does not depend on it. Migrationpaths are unsure.
    """

    entity_type: str = Field(
        ...,
        title="Type of Entity",
        description="Unique name of object within the system",
    )
    entity_id: UUID = Field(
        None,
        title="Identifier of Entity",
        description="Globally unique identifier of this entity",
    )
    entity_meta_summary: str = Field(
        None,
        title="Summary of the subject",
        description="Human readable summary of the content of the object",
    )

    entity_relationships: list = Field(
        [],
        title="Names of attributes containing relationships",
        description="Identifies which attributes relate to other entities",
    )

    entity_meta__fields: list = Field(
        ["entity_meta_summary"],
        title="Names of attributes containing meta fields",
        description="Identifies which attributes contain fields for meta info",
    )

    entity_id__fields: list = Field(
        [],
        title="Names of attributes containing meta fields",
        description="Identifies which attributes are moved to entity_id",
    )

    entity_changelog: list = []

    _event_service = None

    def __init__(self, **kwargs):
        """ Initialized an Entity object

        Initialized an entity object and calls super().__init__ on the pydantic
        based IntrospectableObject model. Allows the setting of _event_service
        for the event_service engine
        """

        super().__init__(**kwargs)
        if "_event_service" in kwargs:
            object.__setattr__(
                self, "_event_service", kwargs["_event_service"]
            )

    def entity_dict(self):
        """ Generates a python dict containing the values of this entity

        Just as the pydantic.dict() method, it returns a dict containing the
        key/values of the attributes of this object.

        The difference is that it will not return private attributes
        (starting with a _) and attributes starting with "entity_". It walks
        recursively over all the related objects.
        """
        rv = {}
        for k, v in dict(self).items():
            if isinstance(v, Entity):
                rv[k] = v.entity_dict()
            elif isinstance(v, IntrospectableObject):
                rv[k] = v.dict()
            else:
                if k.startswith("_") or k.startswith("entity_"):
                    continue

                rv[k] = v

        return rv

    def __setattr__(self, attr, value):
        old_value = self.__getattribute__(attr)

        rv = super().__setattr__(attr, value)

        if not (attr.startswith("_") or attr.startswith("entity_")):
            change = {
                "key": attr,
                "old_value": _reflect(old_value),
                "new_value": _reflect(value),
            }

            self.entity_changelog.append(change)

        return rv

    @staticmethod
    def event(name: str):
        """ New event decorator to capture entity changes and publish events to
        the event_service.

        :param name: Name of event to publish
        :type name: str
        :return: wrapped_entity
        :rtype: func
        """

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            # Clear the changelog when we are not a classmethod
            if isinstance(instance, Entity):
                instance.entity_changelog = []

            rv = wrapped(*args, **kwargs)

            # Check if we are a classmethod instantiating this class. If so,
            # get the changelog from the return value
            changelog = []
            wrapped_entity = None
            if isinstance(instance, Entity):
                changelog = instance.entity_changelog
                wrapped_entity = instance
            elif isinstance(rv, Entity):
                # Classmethod, everything is a change
                dict_of_changes = rv.entity_dict()
                for field in dict_of_changes.keys():
                    value = dict_of_changes[field]

                    changelog.append(
                        {
                            "key": field,
                            "old_value": None,
                            "new_value": _reflect(value),
                        }
                    )
                wrapped_entity = rv

            wrapped_entity._event_service.log_event(
                entity_type=wrapped_entity.__class__.__name__,
                entity_id=str(wrapped_entity.entity_id),
                event_name=name,
                changes=changelog,
                entity_data={},
            )

        return wrapper



import json
from .cqrs import MiddlewareBase
from amqpstorm import Message


def AmqpPublisherMiddleware(
    publisher_name: str, infrastructure_name: str = "amqp"
):
    """Return  `_AMQPPublisherClass` instantiated given params.

    :param publisher_name: name of publisher to get from config
    :type publisher_name: str
    :param infrastructure_name: name of amqp infrastructure, defaults to "amqp"
    :type infrastructure_name: str, optional
    :return: _AMQPPublisher middleware
    :rtype: _AMQPPublisher
    """

    class _AMQPPublisher(MiddlewareBase):
        """Publish all events in the event service to AMQP exchange."""

        def __call__(self, func):
            func()

            self.channel = self.infrastructure_factory.get_infrastructure(
                context=self.context, infrastructure_name=infrastructure_name
            )
            config = self.infrastructure_factory.get_config(
                context=self.context
            )
            publish_settings = config[infrastructure_name]["publish_settings"]
            exchange = publish_settings["exchange"]

            for event in self.event_service.event_list:
                timer = self.statsd.get_timer(event.domain)
                with timer.time(f"publish_amqp_event_time"):
                    properties = {"content_type": "application/json"}
                    event_content = json.dumps(
                        {
                            "id": str(event.uuid),
                            "created_date": event.created_date.isoformat(),
                            "correlation_id": str(self.correlation_id),
                            "context": event.context,
                            "domain": event.domain,
                            "user_uuid": str(event.user_uuid),
                            "entity_type": event.entity_type,
                            "entity_id": str(event.entity_id),
                            "event_name": event.event_name,
                            "changes": event.changes,
                            "entity_data": event.entity_data,
                        },
                        sort_keys=True,
                    )
                    message = Message.create(
                        channel=self.channel,
                        body=event_content,
                        properties=properties,
                    )
                    formatted_domain = str(event.domain).replace(".", "_")
                    routing_key = f"zsnl.v2.{formatted_domain}.{event.entity_type}.{event.event_name}"

                    self.logger.info(
                        f"Publishing event {event.uuid} on {routing_key} - "
                        + str(message.method)
                        + " - "
                        + str(message.properties)
                        + " - EXCHANGE: "
                        + exchange
                    )

                    message.publish(routing_key=routing_key, exchange=exchange)

                self.statsd.get_counter().increment("publish_amqp_event")

    return _AMQPPublisher



import magic
import os
import xml.etree.ElementTree as ET
import zipfile


def is_zip(file_handle):
    file_handle.seek(0, os.SEEK_SET)
    header = file_handle.read(4)

    if header == b"PK\x03\x04":
        return True
    return False


def get_mime_type_from_content(file_handle):
    file_handle.seek(0, os.SEEK_SET)

    # This is needed to resolve an issue with the file handle
    # pointer not being reset despite the seek call
    file_content = file_handle.read(1024 * 1024)
    mime_type = magic.detect_from_content(file_content).mime_type

    return mime_type


def get_mime_type_from_handle(file_handle):
    recognized_office_formats = {
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document.main+xml": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet.main+xml": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        "application/vnd.openxmlformats-officedocument.presentationml.presentation.main+xml": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    }

    mime_type = "application/octet-stream"
    try:
        if is_zip(file_handle):
            zip_handle = zipfile.ZipFile(file_handle)
            xml_content = zip_handle.read("[Content_Types].xml")
            namespaces = {
                "ct": "http://schemas.openxmlformats.org/package/2006/content-types"
            }
            root = ET.fromstring(xml_content)
            content_types = [
                itm.attrib["ContentType"]
                for itm in root.findall("ct:Override", namespaces)
            ]

            for content_type in content_types:
                if content_type in recognized_office_formats:
                    mime_type = recognized_office_formats[content_type]
                    break
        else:
            file_handle.seek(0, os.SEEK_SET)
            mime_type = get_mime_type_from_content(file_handle)
    except (KeyError, zipfile.BadZipFile, ET.ParseError):
        file_handle.seek(0, os.SEEK_SET)
        mime_type = get_mime_type_from_content(file_handle)

    file_handle.seek(0, os.SEEK_SET)
    return mime_type



class CQRSException(Exception):
    """Base exception for things that go wrong in commands and queries"""

    pass


class NotFound(CQRSException):
    """Raised when an item not found.

    This happens when trying to issue a command on a non-existent object
    or querying/retrieving a single non-existent object."""

    pass


class Forbidden(CQRSException):
    """Raised when the user doesn't have permission.

    This can be permission to query an object, or to execute a command."""

    pass


class Conflict(CQRSException):
    """Raised when a command's parameters conflict with current state.

    For instance, when trying to execute a command with a supposedly
    unique identifier that turns out not to be unique."""

    pass


class ConfigurationConflict(CQRSException):
    """Raised when there is a configuration conflict.

    For instance, when trying to retrieve infrastructure from the infrastructure
    factory but the infrastructure can't be initialized because the config is
    missing."""

    pass


class ValidationError(CQRSException):
    """Raised when command parameter validation fails.

    Will contain a list of detected errors."""

    pass



import statsd
import threading
from .. import Base
from ..config.parser import ApacheConfigParser, ConfigParserBase
from ..config.store import FileStore, RedisStore
from ..exceptions import ConfigurationConflict
from collections import namedtuple
from copy import copy
from redis import StrictRedis

CONFIG_STORE_MAP = {"file": FileStore, "redis": RedisStore}
CONFIG_ARGUMENT_MAP = {"redis": StrictRedis}


def _parse_global_config(filename, config_parser: ConfigParserBase):
    """Parse config file with given config_parser.

    :param filename: filename
    :type filename: str
    :param config_parser: config parser to use
    :type config_parser: ConfigParserBase
    :return: config
    :rtype: dict
    """
    with open(filename, "r", encoding="utf-8") as config_file:
        content = config_file.read()

    return config_parser.parse(content)


class InfrastructureFactory(Base):
    """Infrastructure factory class.

    The infrastructure factory will create instances of registered
    "infrastructure" classes, with configuration for a specific context.
    """

    def __init__(
        self, config_file: str, config_parser: ConfigParserBase = None
    ):
        """Initialize an application service factory.

        After reading the configuration, it also configures the defaults
        in the statsd module.

        :param config_file: Global configuration file to read
        :type config_file: str
        :param config_parser: config parser used to parse global configuration
        :type config_parser: ConfigParserBase
        """
        self._local_storage = LocalInfrastructureStorage()
        if config_parser is None:
            config_parser = ApacheConfigParser

        self._config_parser = config_parser()

        self.logger.info(f"Using configuration file '{config_file}'")
        self._global_config = _parse_global_config(
            filename=config_file, config_parser=self._config_parser
        )

        config_store_type = self._global_config["InstanceConfig"]["type"]

        if config_store_type == "none":
            self._instance_config_store = None
        else:
            config_store_args = self._global_config["InstanceConfig"][
                "arguments"
            ]

            try:
                config_store_args = CONFIG_ARGUMENT_MAP[config_store_type](
                    **config_store_args
                )
            except KeyError:
                # There doesn't have to be an argument mapper
                pass

            config_store = CONFIG_STORE_MAP[config_store_type](
                parser=ApacheConfigParser(), arguments=config_store_args
            )

            self._instance_config_store = config_store

        if "statsd" in self._global_config:
            if "disabled" in self._global_config["statsd"]:
                self._global_config["statsd"]["disabled"] = bool(
                    int(self._global_config["statsd"]["disabled"])
                )
            statsd.Connection.set_defaults(**self._global_config["statsd"])
        else:
            # No statsd configuration available; forcefully disable it
            statsd.Connection.set_defaults(disabled=True)

        self._registered_infrastructure = {}

    def __copy__(self):
        """Return a copy of the infrastructure factory.

        The copy will have a new (empty) infrastructure cache and registered
        infrastructure, but will share configuration with the original.
        """

        cls = self.__class__
        copy_factory = cls.__new__(cls)

        copy_factory._local_storage = LocalInfrastructureStorage()
        copy_factory._registered_infrastructure = copy(
            self._registered_infrastructure
        )

        copy_factory._config_parser = self._config_parser
        copy_factory._global_config = self._global_config
        copy_factory._instance_config_store = self._instance_config_store
        return copy_factory

    def get_config(self, context: str) -> dict:
        """Get config from infrafactory for given context.

        :param context: context
        :type context: str
        :return: config
        :rtype: dict
        """
        if self._instance_config_store is None or context is None:
            config = {**self._global_config}
        else:
            instance_config = self._instance_config_store.retrieve(context)
            config = {**self._global_config, **instance_config}
        return config

    def register_infrastructure(self, name: str, infrastructure: object):
        """Register an infrastructure class with the factory.

        :param cls: Class to register in the infrastructure factory
        :type cls: class
        """
        self._registered_infrastructure[name] = infrastructure

    def get_infrastructure(self, context: str, infrastructure_name: str):
        """Retrieve an infrastructure instance for the selected instance.

        If local_storage does not already have the infrastructure instance,
        it will be instantiated from infrastructure factory.

        :param infrastructure_name: Name of the infrastructure class to
            instantiate
        :type infrastructure_name: str
        :return: infrastructure
        :rtype: object
        """
        InfraKey = namedtuple("InfraKey", ["context", "infrastructure_name"])
        infra_key = InfraKey(
            context=context, infrastructure_name=infrastructure_name
        )
        try:
            return self._local_storage.infra[infra_key]
        except KeyError:
            pass

        try:
            infra = self.get_infra_from_infrastructure_factory(
                context=context, infrastructure_name=infrastructure_name
            )
        except KeyError as error:
            raise ConfigurationConflict(
                f"KeyError in infrastructure_factory.get_infrastructure: {error}"
            ) from error

        self._local_storage.infra[infra_key] = infra
        return infra

    def get_infra_from_infrastructure_factory(
        self, context: str, infrastructure_name: str
    ):
        """Retrieve an infrastructure instance for the selected instance.

        :param infrastructure_name: Name of the infrastructure class to
            instantiate
        :type infrastructure_name: str
        :return: infrastructure
        :rtype: object
        """
        config = self.get_config(context=context)
        with self.statsd.get_timer("get_infrastructure").time(
            infrastructure_name
        ):
            infra = self._registered_infrastructure[infrastructure_name](
                config=config
            )
        return infra

    def flush_local_storage(self):
        """Clean up all infrastructure in the local store.

        Usually called at the end of a request cycle. Some infrastructure
        instances might not have the `.clean_up(infrastructure)` method.
        """
        for infra_key, infra in self._local_storage.infra.items():
            try:
                with self.statsd.get_timer(
                    f"{infra_key.infrastructure_name}"
                ).time("clean_up"):
                    self._registered_infrastructure[
                        infra_key.infrastructure_name
                    ].clean_up(infra)
            except AttributeError:
                # Infrastructure does not have clean_up method implemented.
                pass
        self._local_storage.infra = {}
        return


class LocalInfrastructureStorage(threading.local):
    """Thread local storage for infrastructure."""

    def __init__(self):
        """Initilize the local storage for infrastructure with infra dict."""
        self.infra = {}
